class HLISATests(BaseCommand):
    """This command runs HLISA tests"""

    def __init__(self) -> None:
        self.logger = logging.getLogger("openwpm")

    def __repr__(self) -> str:
        return "HLISATestsCommand"

    def execute(
        self,
        webdriver,
        browser_params,
        manager_params,
        extension_socket,
    ) -> None:

        from HLISA.hlisa_action_chains import HLISA_ActionChains

        webdriver.maximize_window()
        webdriver.get('http://localhost:8000/')

        # Clicking test:
        actions = HLISA_ActionChains(webdriver)

        button1 = webdriver.find_element_by_id("button1")
        for i in range(10):
            actions.click(button1)

        # Context clicking test:

        button2 = webdriver.find_element_by_id("button2")
        actions.context_click(button2)

        # Test whether cursor location is correct after scroll:

        actions.scroll_by(0, 100)
        button3 = webdriver.find_element_by_id("button3")
        actions.click(button3)
        
        actions.perform()

        # Test whether the cursor location is correcter after a different page has been loaded:

        webdriver.get('http://localhost:8000/page2.html')

        actions.reset_actions()
        button1 = webdriver.find_element_by_id("button1")
        actions.click(button1)

        # Test drag and drop:

        draggable_element = webdriver.find_element_by_id("draggable_element")
        drag_endpoint = webdriver.find_element_by_id("drag_endpoint")
        actions.drag_and_drop(draggable_element, drag_endpoint)

        # Element with offset test:

        element_with_offset = webdriver.find_element_by_id("element_with_offset")
        actions.move_to_element_with_offset(element_with_offset, 12, 34)
        actions.click()

        # Double click test:

        double_click_element = webdriver.find_element_by_id("double_click")
        actions.double_click(double_click_element)

        # Test drag and drop with offset:

        draggable_element_offset = webdriver.find_element_by_id("draggable_element_offset")
        drag_endpoint_offset = webdriver.find_element_by_id("drag_endpoint_offset")
        actions.drag_and_drop_by_offset(draggable_element_offset, 500, 80)

        # Send keys to element test:

        send_keys_to_element_test = webdriver.find_element_by_id("send_keys_to_element_test")
        actions.send_keys_to_element(send_keys_to_element_test, "test test")

        # Key up and key down test:

        key_up_key_down_test = webdriver.find_element_by_id("key_up_key_down_test")
        actions.key_down("a", key_up_key_down_test)
        actions.key_up("a")
        actions.key_down("b")
        actions.key_up("b")

        # End of tests

        actions.pause(50)
        actions.perform()


import math
import time
import random
import numpy as np

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.common.actions.pointer_input import PointerInput
from selenium.webdriver.remote.webelement import WebElement

class HL_Util:
    # (normal distribution)
    # Takes an element and returns coordinates somewhere in the element. If the element is not visable, it returns 0.
    def behavorial_element_coordinates(self, webdriver, element):
        x_relative = int(element.rect['x']) - webdriver.execute_script("return window.pageXOffset;")
        y_relative = int(element.rect['y']) - webdriver.execute_script("return window.pageYOffset;")
        viewport_width = webdriver.execute_script("return window.innerWidth")
        viewport_height = webdriver.execute_script("return window.innerHeight")
        counter = 0
        for i in range(100): # Try 10 random positions, as some positions are not in round buttons.
            x = x_relative + int(np.random.normal(int(element.rect['width']*0.5), int(element.rect['width']*0.2)))
            y = y_relative + int(np.random.normal(int(element.rect['height']*0.5), int(element.rect['height']*0.2)))
            coords_in_button = webdriver.execute_script("return document.elementFromPoint(" + str(x) + ", " + str(y) + ") === arguments[0];", element)
            if x < 0 or y < 0 or x > viewport_width or y > viewport_height:
                coords_in_button = False # If the element is partly in the viewport, a part of the element is outside of it. In that case, try again. This is not the best solution (non-deterministic), it would be better to limit the sample space.
            if coords_in_button:
                return (x, y)
        return None

    # Returns a number from a normal distribution that is larger or equal to parameter 'minimal'.
    # Due to the minimum value, the returned values will not form a normal distribution.
    # To minimize this effect, values that would have been smaller than the minimum are not drawn
    # again, but get added a random small value. The new number will never become larger as the mean.
    def std_positive(mean, std, minimal):
        sample = np.random.normal(mean, std)
        while sample < minimal:
            sample += random.random() * (mean - minimal)
        return sample

    # The function to replace the original Selenium function that does not support specifying the duration.
    def create_pointer_move(self, duration=50, x=None, y=None, origin=None):
        action = dict(type="pointerMove", duration=duration)
        action["x"] = x
        action["y"] = y
        if isinstance(origin, WebElement):
            action["origin"] = {"element-6066-11e4-a52e-4f735466cecf": origin.id}
        elif origin is not None:
            action["origin"] = origin

        self.add_action(action)

    # Replace a function in the original Selenium API to increase mouse movement speed.
    def increaseMousemovementSpeed():
        PointerInput.create_pointer_move = HL_Util.create_pointer_move

    



import math
import time
import random
import logging
import numpy as np

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By

from HLISA.util import HL_Util
from HLISA.selenium_actions import HL_Selenium_Actions

class HL_Additional_Actions:
    scroll_tick_size = 57

    def __init__(self, webdriver):
        self.webdriver = webdriver

    # Function that defines a short pause between actions
    def shortPauze(self):
        time.sleep(random.random() + 0.5)

    # First scrolls to get the element into the viewport, then performs the movement
    def move_to_element_outside_viewport(self, element, addDelayAfter=True):
        viewport_height = self.webdriver.execute_script("return window.innerHeight")
        y_relative = int(element.rect['y']) - self.webdriver.execute_script("return window.pageYOffset;")
        if y_relative < 0:
            self.scroll_by(0, y_relative)
        elif y_relative > viewport_height:
            self.scroll_by(0, y_relative - viewport_height/2)
        x, y = HL_Util.behavorial_element_coordinates("", self.webdriver, element)
        selenium_actions = HL_Selenium_Actions(self.webdriver)
        selenium_actions.move_to(x, y, addDelayAfter)
        selenium_actions.perform()

    # This function scrolls a few pixels further if the parameter is not a multiple of a standard scroll value.
    # It would be detectable otherwise.
    def scroll_by(self, x_diff, y_diff, addDelayAfter=True):    
        if x_diff != 0:
            logging.critical("Scrolling horizontal not implemented")
        self.scroll_vertical(y_diff)
        if addDelayAfter:
            self.shortPauze()

    def scroll_vertical(self, y_diff):
        scroll_ticks = 0
        current_y = self.webdriver.execute_script("return window.pageYOffset;")
        if y_diff > 0:
            max_y = self.webdriver.execute_script("return Math.max(document.body.scrollHeight, document.body.offsetHeight, document.documentElement.clientHeight, document.documentElement.scrollHeight, document.documentElement.offsetHeight);")
            y_diff = min(y_diff, max_y - current_y) # Prevent scrolling too far
            while y_diff > 0:
                y_diff = self.scroll_tick(self.scroll_tick_size, scroll_ticks, y_diff)
                scroll_ticks += 1
        else:
            min_y = 0
            y_diff = max(y_diff, min_y - current_y) # Prevent scrolling too far
            while y_diff < 0:
                y_diff = self.scroll_tick((-1 * self.scroll_tick_size), scroll_ticks, y_diff)
                scroll_ticks += 1
        new_y = self.webdriver.execute_script("return window.pageYOffset;")
        scrolled_distance = abs(current_y - new_y)
        if scrolled_distance <= abs(y_diff) - self.scroll_tick_size:
            if y_diff >= 0:
                self.scroll_vertical(y_diff - scrolled_distance)
            else:
                self.scroll_vertical(y_diff + scrolled_distance)

    # Scrolls one tick
    def scroll_tick(self, pixelAmount, scroll_ticks, y_diff):
        self.webdriver.execute_script("window.scrollBy(0, " + str(pixelAmount) + ")")
        y_diff -= pixelAmount
        time.sleep(0.05 + (random.random()/200))
        if scroll_ticks % 7 == 0:
            time.sleep(HL_Util.std_positive(0.5, 0.1, 0))
        return y_diff

    # This function scrolls a few pixels further if the parameter is not a multiple of a standard scroll value.
    # It would be detectable otherwise.
    def scroll_to(self, x, y, addDelayAfter=True):
        self.shortPauze()  
        current_x = self.webdriver.execute_script("return window.pageXOffset;")
        if current_x != x:
            logger.error("Scrolling horizontal not yet implemented")
        current_y = self.webdriver.execute_script("return window.pageYOffset;")
        y_diff = y - current_y
        self.scroll_by(x, y_diff, addDelayAfter)



import math
import time
import random
import numpy as np

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains

from HLISA.util import HL_Util
from HLISA.selenium_actions import HL_Selenium_Actions
from HLISA.additional_actions import HL_Additional_Actions

# This object holds its own chain of actions in self.chain.
# Every API call on this object adds the action to the chain,
# and if the action is ActionChain based itself (meaning it 
# only executes after a .perform()), the .perform() is also 
# called.
class HLISA_ActionChains:

    def __init__(self, webdriver, browser_resets_cursor_location=True):
        self.webdriver = webdriver
        self.chain = []
        HL_Selenium_Actions.browser_resets_cursor_location = browser_resets_cursor_location

    ##### Standard Selenium action chain methods #####

    def click(self, element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.click(element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def click_and_hold(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.click_and_hold(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def context_click(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.context_click(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def double_click(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.double_click(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def drag_and_drop(self, source, target, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.drag_and_drop(source, target, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def drag_and_drop_by_offset(self, source, xoffset, yoffset, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.drag_and_drop_by_offset(source, xoffset, yoffset, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def key_down(self, value, element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.key_down(value, element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def key_up(self, value, element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.key_up(value, element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_by_offset(self, x, y, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_by_offset(x, y, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_to_element(self, element, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_to_element(element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_to_element_with_offset(self, to_element, xoffset, yoffset, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_to_element_with_offset(to_element, xoffset, yoffset, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def pause(self, seconds):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.pause(seconds))
        self.chain.append(lambda: actions.perform())
        return self

    def perform(self):
        for action in self.chain:
            action()
        if HL_Selenium_Actions.selenium_version >= 4:
            self.reset_actions()
        return self

    def release(self, on_element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.release(on_element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def reset_actions(self):
        self.chain = []
        return self

    def send_keys(self, keys_to_send, element=None, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.send_keys(keys_to_send, element, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def send_keys_to_element(self, element, keys_to_send, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.send_keys_to_element(element, keys_to_send, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    ##### Additional actions #####

    def move_to(self, x, y, addDelayAfter=True):
        actions = HL_Selenium_Actions(self.webdriver)
        self.chain.append(lambda: actions.move_to(x, y, addDelayAfter))
        self.chain.append(lambda: actions.perform())
        return self

    def move_to_element_outside_viewport(self, element, addDelayAfter=True):
        additional_actions = HL_Additional_Actions(self.webdriver)
        self.chain.append(lambda: additional_actions.move_to_element_outside_viewport(element, addDelayAfter))
        return self

    def scroll_by(self, x_diff, y_diff, addDelayAfter=True):
        additional_actions = HL_Additional_Actions(self.webdriver)
        self.chain.append(lambda: additional_actions.scroll_by(x_diff, y_diff, addDelayAfter))
        return self

    def scroll_to(self, x_diff, y_diff, addDelayAfter=True):
        additional_actions = HL_Additional_Actions(self.webdriver)
        self.chain.append(lambda: additional_actions.scroll_to(x_diff, y_diff, addDelayAfter))
        return self

    def back(self):
        HL_Selenium_Actions.page_identifier = str(self.webdriver.execute_script("return window.performance.timing.domContentLoadedEventEnd"))
        self.webdriver.back()
        return self

    def forward(self):
        HL_Selenium_Actions.page_identifier = str(self.webdriver.execute_script("return window.performance.timing.domContentLoadedEventEnd"))
        self.webdriver.forward()
        return self
